"""Module containing tests for the selenium utility functions"""
import os
from datetime import datetime, timedelta
from typing import Tuple, Any
from unittest import TestCase, main
from unittest.mock import Mock, patch

from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By

from judah.utils.selenium import (
    get_web_driver,
    get_html_element_by_xpath,
    visit_website,
    wait_for_download_to_complete,
    wait_for_element_to_load,
    WebDriverOptions
)

_PARENT_FOLDER = os.path.dirname(__file__)
_MOCK_ASSET_FOLDER_PATH = os.path.join(os.path.dirname(_PARENT_FOLDER), 'assets')


class TestSeleniumUtilities(TestCase):
    """Tests for the selenium utility functions"""

    def test_get_web_driver(self):
        """
        Should return a working instance of a chrome webdriver
        FIXME: Unfortunately, I am unable to check if any of the options have been applied appropriately
        """
        download_location = os.path.join(_MOCK_ASSET_FOLDER_PATH, 'csv')
        options = WebDriverOptions(downloads_folder_location=download_location)

        driver = get_web_driver(options)
        self.assertIsInstance(driver, webdriver.Chrome)

        driver.close()
        driver.quit()

    def test_visit_website(self):
        """
        Should make a GET request to the URL passed to it
        """
        mock_driver = Mock(spec=webdriver.Chrome)
        mock_driver.get = Mock()
        url = 'http://example.com'

        visit_website(mock_driver, website_url=url)

        mock_driver.get.assert_called_once_with(url=url)

    def test_get_html_element_by_xpath(self):
        """
        Should get the html element by given xpath, waiting for it only up to a given timeout
        and giving up
        """
        start_time = datetime.now()
        mock_driver = Mock(spec=webdriver.Chrome)
        timeout_in_seconds = 2
        seconds_before_element_shows_up = 1
        test_xpath = 'hello_world'
        test_element = {'foo': 'bar'}

        def mock_find_element_by_xpath(xpath: str):
            """Waits for a given time before returning an element"""
            if datetime.now() > start_time + timedelta(seconds=seconds_before_element_shows_up):
                return test_element

            return False

        mock_driver.find_element_by_xpath = Mock()
        mock_driver.find_element_by_xpath.side_effect = mock_find_element_by_xpath

        element = get_html_element_by_xpath(driver=mock_driver, xpath=test_xpath, timeout=timeout_in_seconds)

        mock_driver.find_element_by_xpath.assert_called_with(xpath=test_xpath)
        self.assertEqual(element, test_element)

    def test_timed_out_get_html_element_by_xpath(self):
        """
        Should raise TimeoutException if the element takes longer than the timeout setting to show up
        """
        mock_driver = Mock(spec=webdriver.Chrome)
        timeout_in_seconds = 2
        test_xpath = 'hello_world'

        mock_driver.find_element_by_xpath.return_value = False
        self.assertRaises(TimeoutException, get_html_element_by_xpath, driver=mock_driver, xpath=test_xpath,
                          timeout=timeout_in_seconds)

    @patch('os.path.isfile')
    @patch('os.path.exists')
    def test_wait_for_download_to_complete(self, mock_os_path_exists, mock_os_path_is_file):
        """
        Should wait for download to complete before returning
        """
        expected_file_path = 'dummy.csv'
        timeout = 2
        seconds_before_download_completes = 1
        mock_os_path_is_file.return_value = True
        start_time = datetime.now()

        def mock_os_path_exists_side_effect(file_name: str):
            """Waits for a given time before returning True"""
            if datetime.now() > start_time + timedelta(seconds=seconds_before_download_completes):
                return True

            return False

        mock_os_path_exists.side_effect = mock_os_path_exists_side_effect

        self.assertLessEqual((datetime.now() - start_time).total_seconds(), seconds_before_download_completes)

        wait_for_download_to_complete(expected_file_path=expected_file_path, timeout=timeout)

        self.assertGreater((datetime.now() - start_time).total_seconds(), seconds_before_download_completes)

    @patch('os.path.isfile')
    @patch('os.path.exists')
    def test_wait_for_download_to_complete_not_file(self, mock_os_path_exists, mock_os_path_is_file):
        """
        Should raise IOError if the file is not an actual file
        """
        expected_file_path = 'dummy.csv'
        timeout = 2
        mock_os_path_exists.return_value = True
        mock_os_path_is_file.return_value = False

        self.assertRaises(IOError, wait_for_download_to_complete,
                          expected_file_path=expected_file_path, timeout=timeout)

    @patch('judah.utils.selenium.EC')
    def test_wait_for_element_to_load(self, mock_ec):
        """
        Should wait for the element to load before returning
        """
        start_time = datetime.now()
        mock_driver = Mock(spec=webdriver.Chrome)
        timeout_in_seconds = 2
        seconds_before_element_shows_up = 1
        test_xpath = 'hello_world'
        test_element = {'foo': 'bar'}

        def mock_presence_of_element_located(search_item: Tuple[Any]):
            """Waits for a given time before returning an element"""
            if datetime.now() > start_time + timedelta(seconds=seconds_before_element_shows_up):
                return test_element

            return False

        mock_ec.presence_of_element_located = Mock()
        mock_ec.presence_of_element_located.return_value = mock_presence_of_element_located

        self.assertLessEqual((datetime.now() - start_time).total_seconds(), seconds_before_element_shows_up)

        wait_for_element_to_load(driver=mock_driver, xpath=test_xpath, timeout=timeout_in_seconds)

        self.assertGreater((datetime.now() - start_time).total_seconds(), seconds_before_element_shows_up)
        mock_ec.presence_of_element_located.assert_called_once_with((By.XPATH, test_xpath))


if __name__ == '__main__':
    main()



"""Module containing the utilities that are to do with CRON jobs"""
from datetime import date, time, datetime, timedelta
from typing import Optional, Dict, Any

import pytz
from pydantic import BaseModel
from enum import IntEnum

from judah.utils.dates import add_month_to_datetime_but_maintain_day


class Month(IntEnum):
    JANUARY = 1
    FEBRUARY = 2
    MARCH = 3
    APRIL = 4
    MAY = 5
    JUNE = 6
    JULY = 7
    AUGUST = 8
    SEPTEMBER = 9
    OCTOBER = 10
    NOVEMBER = 11
    DECEMBER = 12


class WeekDay(IntEnum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6


class Hour(IntEnum):
    TWELVE_AM = 0
    ONE_AM = 1
    TWO_AM = 2
    THREE_AM = 3
    FOUR_AM = 4
    FIVE_AM = 5
    SIX_AM = 6
    SEVEN_AM = 7
    EIGHT_AM = 8
    NINE_AM = 9
    TEN_AM = 10
    ELEVEN_AM = 11
    TWELVE_PM = 12
    ONE_PM = 13
    TWO_PM = 14
    THREE_PM = 15
    FOUR_PM = 16
    FIVE_PM = 17
    SIX_PM = 18
    SEVEN_PM = 19
    EIGHT_PM = 20
    NINE_PM = 21
    TEN_PM = 22
    ELEVEN_PM = 23


class CronType(BaseModel):
    day_of_week: Optional[WeekDay] = None
    year: Optional[int] = None
    month: Optional[Month] = None
    day: Optional[int] = None
    hour: Optional[Hour] = None
    minute: Optional[int] = None
    second: Optional[int] = None
    timezone_string: Optional[str] = 'UTC'

    class Config:
        arbitrary_types_allowed = True

    def get_next_utc_timestamp(self) -> Optional[datetime]:
        """
        The main function that is called by the Controller to get the next UTC timestamp
        from which to extract the appropriate delay that the controller will have to wait for
        """
        utc_now = datetime.now(tz=pytz.UTC)
        generated_timestamp = self.__get_timestamp_from_fields(default_timestamp=utc_now)

        if generated_timestamp >= utc_now:
            next_timestamp = generated_timestamp
        else:
            next_timestamp = self.__get_next_timestamp(generated_timestamp)

        return next_timestamp if next_timestamp >= utc_now else None

    def __update_fields_with_defaults(self):
        """Updates the fields with the generated default values"""
        defaults = self._generate_field_defaults()

        for key, value in defaults.items():
            setattr(self, key, value)

    def __get_timestamp_from_fields(self, default_timestamp: datetime):
        """Constructs a timestamp from the fields, replacing None values with those from the default timestamp"""
        self.__update_fields_with_defaults()

        next_date = self._get_date()
        next_time = self._get_time()
        timezone = pytz.timezone(self.timezone_string)

        datetime_options = dict(
            year=default_timestamp.year, month=default_timestamp.month, day=default_timestamp.day,
            hour=default_timestamp.hour, minute=default_timestamp.minute, second=default_timestamp.second)

        if next_date is not None:
            datetime_options = {
                **datetime_options, **dict(day=next_date.day, year=next_date.year, month=next_date.month)}

        if next_time is not None:
            datetime_options = {
                **datetime_options, **dict(hour=next_time.hour, minute=next_time.minute, second=next_time.second)}

        return datetime(**datetime_options, tzinfo=timezone).astimezone(pytz.UTC)

    def __get_next_timestamp(self, current_timestamp: datetime):
        """Gets the timestamp after the current timestamp"""
        if isinstance(self.year, int):
            # once a year is set, only one timestamp can be derived from it.
            return current_timestamp
        elif isinstance(self.day_of_week, WeekDay):
            # shift it by 7 days
            return current_timestamp + timedelta(days=7)
        elif isinstance(self.month, Month):
            # forward it by a year
            return current_timestamp.replace(year=current_timestamp.year + 1)
        elif isinstance(self.day, int):
            # forward it by a month
            return add_month_to_datetime_but_maintain_day(current_timestamp)
        elif isinstance(self.hour, Hour):
            # forward it by a day
            return current_timestamp + timedelta(days=1)
        elif isinstance(self.minute, int):
            # forward it by an hour
            return current_timestamp + timedelta(hours=1)
        elif isinstance(self.second, int):
            # forward it by a minute
            return current_timestamp + timedelta(minutes=1)
        return current_timestamp

    def _get_next_date_from_day_of_week(self) -> Optional[date]:
        """
        Constructs the date of the next day_of_week e.g. next Sunday if day_of_week = WeekDay.SUNDAY
        """
        if self.day_of_week is None:
            return None

        today = date.today()

        current_weekday = today.weekday()
        number_of_days_ahead = self.day_of_week.value - current_weekday

        if number_of_days_ahead < 0:
            # the self.day_of_week has already passed for current week so go to next week
            number_of_days_ahead += 7

        return today + timedelta(days=number_of_days_ahead)

    def _get_date(self) -> Optional[date]:
        """
        Constructs the datetime.date basing on the ‘day_of_week’, ‘year’, ‘month’ and ‘day’
        """
        next_date_for_day_of_week = self._get_next_date_from_day_of_week()

        if isinstance(next_date_for_day_of_week, date):
            return next_date_for_day_of_week

        today = date.today()
        date_fields = ('day', 'year',)
        date_options = {key: getattr(self, key) for key in date_fields if getattr(self, key) is not None}

        if self.month is not None:
            date_options['month'] = self.month.value

        if len(date_options) == 0:
            return None

        # fill up the date options with the defaults
        date_options = {'day': today.day, 'year': today.year, 'month': Month(today.month), **date_options}

        return date(**date_options)

    def _get_time(self) -> Optional[time]:
        """
        Constructs the datetime.time basing on the ‘hour’, ‘minute’, ‘second’, 
        """
        time_fields = ('minute', 'second',)
        time_options = {key: getattr(self, key) for key in time_fields if getattr(self, key) is not None}
        utc_now = datetime.now(tz=pytz.UTC)

        if self.hour is not None:
            time_options['hour'] = self.hour.value

        if len(time_options) == 0:
            return None

        # fill up the time options with the defaults
        time_options = {'hour': utc_now.hour, 'minute': utc_now.minute, 'second': utc_now.second, **time_options}

        return time(**time_options)

    def _generate_field_defaults(self) -> Dict[str, Any]:
        """
        Generates the default values of the other fields if a particular field is set
        e.g. if the ‘hour’ field is not None, the ‘minute’, ‘second’,  fields
        will automatically be set to zero (0)
        """
        defaults = self.dict().copy()
        date_fields = ('day_of_week', 'year', 'month', 'day',)

        should_set_month = False
        should_set_day = False
        should_set_hour = False
        should_set_minute = False
        should_set_second = False

        if defaults['year'] is not None:
            should_set_month = defaults['month'] is None
            should_set_day = defaults['day'] is None
        if defaults['month'] is not None:
            should_set_day = defaults['day'] is None
        if any([defaults[field] is not None for field in date_fields]):
            should_set_hour = defaults['hour'] is None
            should_set_minute = defaults['minute'] is None
            should_set_second = defaults['second'] is None
        if self.hour is not None:
            should_set_minute = defaults['minute'] is None
            should_set_second = defaults['second'] is None
        if self.minute is not None:
            should_set_second = defaults['second'] is None

        if should_set_month:
            defaults['month'] = Month.JANUARY
        if should_set_day:
            defaults['day'] = 1
        if should_set_hour:
            defaults['hour'] = Hour.TWELVE_AM
        if should_set_minute:
            defaults['minute'] = 0
        if should_set_second:
            defaults['second'] = 0

        return defaults



from datetime import date
from typing import Optional, Type, Iterator, Dict, Any

from ..base import BaseController
from ...sources.base.date_based import DateBasedBaseSource


class DateBasedBaseController(BaseController):
    """Base controller class for date based data sources"""
    _source: DateBasedBaseSource
    _last_timestamp: Optional[date] = None

    start_date: Optional[date] = None
    end_date: Optional[date] = None
    source_class: Type[DateBasedBaseSource]

    @classmethod
    def _update_last_timestamp(cls):
        """Updates the last timestamp property of this class"""
        super()._update_last_timestamp()

        if isinstance(cls._last_timestamp, date):
            cls.start_date = cls._last_timestamp

    @classmethod
    def _query_source(cls, ) -> Iterator[Dict[Any, Any]]:
        """Queries the data source"""
        cls._update_source(start_date=cls.start_date, end_date=cls.end_date)
        return cls._source.get()



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="judah",
    version="0.0.3",
    description="A simple service-oriented ETL framework for integrations",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/sopherapps/judah",
    author="Martin Ahindura",
    author_email="team.sopherapps@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(exclude=("test",)),
    include_package_data=True,
    install_requires=[
        "pydantic",
        "python-dotenv",
        "email-notifier",
        "psycopg2-binary",
        "selenium",
        "requests",
        "SQLAlchemy",
        "xml-stream",
        "webdriver-manager",
        "bonobo",
        "xlrd"
    ],
    entry_points={
    },
)



import os
from datetime import date, timedelta, datetime
from typing import Optional, Tuple, Union

import pytz


def convert_date_to_string(
        date_value: date, date_format: str = '%Y-%m-%dT%T%ZZ', timezone_separator: Optional[str] = None,
        timezone_name: Optional[str] = None) -> str:
    """Converts a date to the expected string"""
    timezone = pytz.utc if timezone_name is None else pytz.timezone(timezone_name)

    datetime_value = datetime(
        year=date_value.year, month=date_value.month, day=date_value.day, tzinfo=pytz.utc).astimezone(
        timezone)
    # remove any time shift due to timezones
    datetime_value = datetime_value - datetime_value.utcoffset()
    formatted_datetime_string = datetime_value.strftime(date_format)

    if isinstance(timezone_separator, str):
        offset_hours = formatted_datetime_string[:-2]
        offset_minutes = formatted_datetime_string[-2:]

        formatted_datetime_string = f'{offset_hours}{timezone_separator}{offset_minutes}'

    return formatted_datetime_string


def get_default_start_year() -> Optional[int]:
    """Returns the default year"""
    start_year = os.getenv('HISTORICAL_STARTING_YEAR', None)

    if start_year is None:
        return None

    return int(start_year)


def get_default_start_month() -> Optional[int]:
    """Returns the default start month"""
    start_month = os.getenv('HISTORICAL_STARTING_MONTH', None)

    if start_month is None:
        return None

    return int(start_month)


def get_quarter_from_month(month: int):
    """Returns the quarter for a given month"""
    month_quarter_map = {1: 1, 2: 1, 3: 1, 4: 2, 5: 2, 6: 2, 7: 3, 8: 3, 9: 3, 10: 4, 11: 4, 12: 4}
    return month_quarter_map[month]


def get_default_start_quarter() -> Optional[int]:
    """Returns the default start quarter"""
    start_month = get_default_start_month()

    if start_month is None:
        return None

    return get_quarter_from_month(month=start_month)


def get_current_year() -> int:
    """Get's the current year basing on the system time"""
    return datetime.now().year


def get_current_quarter() -> int:
    """Get's the current quarter basing on the system time"""
    return get_quarter_from_month(month=datetime.now().month)


def get_default_historical_start_date() -> Optional[date]:
    """Extracts the default start_date from the environment"""
    start_year = get_default_start_year()
    start_month = int(os.getenv('HISTORICAL_STARTING_MONTH', 1))
    start_day = int(os.getenv('HISTORICAL_STARTING_DAY', 1))

    if start_year is None:
        return None

    return date(year=start_year, month=start_month, day=start_day)


def get_default_historical_start_quarter_and_year() -> Optional[Tuple[int, int]]:
    """Extracts the default start_date from the environment"""
    start_year = get_default_start_year()
    start_quarter = get_default_start_quarter()

    if start_year is None:
        return None

    if start_quarter is None:
        start_quarter = 1

    return start_quarter, start_year


def update_date(date_to_update: date, days_to_increment_by: int, days_to_decrement_by: int):
    """Updates a given date by incrementing and/or decrementing it"""
    return date_to_update + timedelta(days=days_to_increment_by) - timedelta(days=days_to_decrement_by)


def update_quarter_year_tuple(
        quarter_year_tuple: Tuple[int, int],
        quarters_to_increment_by: int,
        quarters_to_decrement_by: int) -> Tuple[int, int]:
    """Updates a given quarter_year_tuple by incrementing and/or decrementing it"""
    quarter, year = quarter_year_tuple
    net_quarters_delta = quarters_to_increment_by - quarters_to_decrement_by
    number_of_quarters_per_year = 4

    # convert (quarter, year) tuple to quarters since year 1, quarter 1
    total_quarters_since_year_1_quarter_1 = (quarter - 1) + ((year - 1) * number_of_quarters_per_year)
    new_total_quarters_since_year_1_quarter_1 = total_quarters_since_year_1_quarter_1 + net_quarters_delta

    years_after_year_1 = int(new_total_quarters_since_year_1_quarter_1 / number_of_quarters_per_year)
    updated_year = 1 + years_after_year_1

    quarters_after_quarter_1 = new_total_quarters_since_year_1_quarter_1 % number_of_quarters_per_year
    updated_quarter = 1 + quarters_after_quarter_1

    return updated_quarter, updated_year


def convert_date_to_quarter_year_tuple(date_value: date) -> Tuple[int, int]:
    """Converts a given date to a quarter_year tuple"""
    year = date_value.year
    month = date_value.month
    quarter = get_quarter_from_month(month=month)
    return quarter, year


def change_datetime_format(datetime_value: str, old: str, new: str):
    """Converts a date string's format to another format"""
    return datetime.strptime(datetime_value, old).strftime(new)


def add_month_to_datetime_but_maintain_day(datetime_value: Union[datetime, date]) -> Optional[Union[date, datetime]]:
    """Adds a month to the given date or datetime, maintaining the day itself"""
    try:
        return datetime_value.replace(month=datetime_value.month + 1)
    except ValueError:
        if datetime_value.month == 12:
            return datetime_value.replace(year=datetime_value.year + 1, month=1)
    return None



"""
Module containing tests for the DatabaseConnection class in the connection module
in the database destination package
"""
import os
from unittest import TestCase, main
from unittest.mock import Mock

from sqlalchemy.orm import scoped_session

from judah.destinations.database.connection import DatabaseConnection
from judah.destinations.database.config import DatabaseConnectionConfig

_PARENT_FOLDER = os.path.dirname(__file__)
_MOCK_ASSET_FOLDER_PATH = os.path.join(os.path.dirname(os.path.dirname(_PARENT_FOLDER)), 'assets')


class TestDatabaseConnection(TestCase):
    """Tests for the DatabaseConnection base class"""

    def setUp(self) -> None:
        """Initialize a few variables"""
        self.test_db_1_uri = f'sqlite:///{_MOCK_ASSET_FOLDER_PATH}/test_db_1'
        self.test_db_2_uri = f'sqlite:///{_MOCK_ASSET_FOLDER_PATH}/test_db_2'
        self.test_db_1_config = DatabaseConnectionConfig(db_uri=self.test_db_1_uri)
        self.test_db_2_config = DatabaseConnectionConfig(db_uri=self.test_db_2_uri)

    def test_context_manager_enter(self):
        """On entering the context manager, a scoped session is created and saved to the db_session attribute"""
        test_db_1_connection = DatabaseConnection(self.test_db_1_config)
        self.assertIsNone(test_db_1_connection.db_session)

        with DatabaseConnection(self.test_db_2_config) as test_db_2_connection:
            self.assertIsInstance(test_db_2_connection.db_session, scoped_session)

    def test_context_manager_exit(self):
        """On exiting the context manager, the db_session is closed"""
        with DatabaseConnection(self.test_db_1_config) as test_db_1_connection:
            # create a dummy session to add to the scoped_session's registry
            dummy_session = test_db_1_connection.db_session()
            self.assertTrue(test_db_1_connection.db_session.registry.has())

        self.assertFalse(test_db_1_connection.db_session.registry.has())

    def test_context_manager_exit_with_no_autocommit(self):
        """
        On exiting the context manager when self.db_connection_config.autocommit is False,
        the changes are committed before the session is closed when exiting the context manager.
        """
        db_config_with_no_autocommit = DatabaseConnectionConfig(db_uri=self.test_db_1_uri, autocommit=False)
        with DatabaseConnection(db_config_with_no_autocommit) as db_connection:
            db_connection.db_session.commit = Mock()

        db_connection.db_session.commit.assert_called_once()

    def test_context_manager_exit_with_autocommit(self):
        """
        On exiting the context manager when self.db_connection_config.autocommit is True,
        the changes are not committed before the session is closed when exiting the context manager.
        """
        db_config_with_autocommit = DatabaseConnectionConfig(db_uri=self.test_db_1_uri, autocommit=True)
        with DatabaseConnection(db_config_with_autocommit) as db_connection:
            db_connection.db_session.commit = Mock()

        db_connection.db_session.commit.assert_not_called()

    def test_execute_sql(self):
        """Should execute the sql passed as a parameter"""
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS dummy ( id INTEGER PRIMARY KEY AUTOINCREMENT,
        name TEXT)
        """
        delete_sql = """
        DELETE FROM dummy WHERE name = :name
        """
        insert_sql = """
        INSERT INTO dummy (name) VALUES (:name)
        """
        select_sql = """
        SELECT * FROM dummy
        """
        test_name = 'John Doe'

        with DatabaseConnection(self.test_db_1_config) as db_connection:
            db_connection.execute_sql(create_table_sql)
            db_connection.execute_sql(insert_sql, {'name': test_name})
            names = [item.name for item in db_connection.execute_sql(select_sql).fetchmany()]
            self.assertEqual(names[0], test_name)
            db_connection.execute_sql(delete_sql, {'name': test_name})

    def test_get_db_connection(self):
        """Should get the already initialized instance of DatabaseConnection with the given db_uri"""
        test_db_connection = DatabaseConnection(self.test_db_1_config)
        self.assertEqual(test_db_connection, DatabaseConnection.get_db_connection(self.test_db_1_config))

    def test_close(self):
        """Closes the db connection engine"""
        db_connection = DatabaseConnection(self.test_db_1_config)
        old_dispose = db_connection.connection_engine.dispose
        db_connection.connection_engine.dispose = Mock()
        db_connection.connection_engine.dispose.side_effect = old_dispose

        db_connection.close()
        db_connection.connection_engine.dispose.assert_called_once()

    def test_close_all_connections(self):
        """Should close all connections in the dictionary of connections"""
        db_connection_1 = DatabaseConnection(self.test_db_1_config)
        db_connection_2 = DatabaseConnection(self.test_db_2_config)

        old_dispose_1 = db_connection_1.connection_engine.dispose
        db_connection_1.connection_engine.dispose = Mock()
        db_connection_1.connection_engine.dispose.side_effect = old_dispose_1

        old_dispose_2 = db_connection_2.connection_engine.dispose
        db_connection_2.connection_engine.dispose = Mock()
        db_connection_2.connection_engine.dispose.side_effect = old_dispose_2

        DatabaseConnection.close_all_connections()
        db_connection_1.connection_engine.dispose.assert_called_once()
        db_connection_2.connection_engine.dispose.assert_called_once()

    def test_open_connections(self):
        """Should open the database connections whose configs have been passed to it"""
        self.assertDictEqual(DatabaseConnection.connections, {})
        DatabaseConnection.open_connections([self.test_db_1_config, self.test_db_2_config])

        self.assertEqual(len(DatabaseConnection.connections), 2)
        self.assertIsInstance(DatabaseConnection.connections[self.test_db_1_uri], DatabaseConnection)
        self.assertIsInstance(DatabaseConnection.connections[self.test_db_2_uri], DatabaseConnection)

    def test_remove_connection(self):
        """Should remove a database connection of a given db_uri from the dictionary of connections"""
        db_connection_1 = DatabaseConnection(self.test_db_1_config)
        db_connection_2 = DatabaseConnection(self.test_db_2_config)

        self.assertIsInstance(DatabaseConnection.connections[self.test_db_1_uri], DatabaseConnection)

        DatabaseConnection.remove_connection(self.test_db_1_uri)

        self.assertIsNone(DatabaseConnection.connections.get(self.test_db_1_uri, None))
        self.assertIsInstance(DatabaseConnection.connections[self.test_db_2_uri], DatabaseConnection)

    def test_remove_all_connections(self):
        """Should remove all database connections in the dictionary of connections"""
        db_connection_1 = DatabaseConnection(self.test_db_1_config)
        db_connection_2 = DatabaseConnection(self.test_db_2_config)

        self.assertIsInstance(DatabaseConnection.connections[self.test_db_1_uri], DatabaseConnection)

        DatabaseConnection.remove_all_connections()

        self.assertIsNone(DatabaseConnection.connections.get(self.test_db_1_uri, None))
        self.assertIsNone(DatabaseConnection.connections.get(self.test_db_2_uri, None))

    def tearDown(self) -> None:
        DatabaseConnection.close_all_connections()
        DatabaseConnection.remove_all_connections()


if __name__ == '__main__':
    main()



"""Module containing tests for the cron utility functions"""
from datetime import time, date, timedelta, datetime
from unittest import TestCase, main
from unittest.mock import patch

import pytz
from freezegun import freeze_time

from judah.utils.cron import CronType, WeekDay, Hour, Month


class TestCronType(TestCase):
    """Tests for the CronType class"""

    def setUp(self) -> None:
        """Initialize some variables"""
        self.timezone_name = "Europe/Paris"

    def test_generate_field_defaults(self):
        """
        Should return a dictionary of the default values of the other fields if a particular field is set
        """
        cron_job = CronType()
        expected_defaults = dict(
            day_of_week=None, year=None, month=None, day=None, hour=None,
            minute=None, second=None, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_day_of_week_is_set(self):
        """
        Should return a dictionary with hour set to Hour.TWELVE_AM, and minute, second, millisecond set to 0
        and day_of_week set to that given day_of_week
        """
        cron_job = CronType(day_of_week=WeekDay.THURSDAY)
        expected_defaults = dict(
            day_of_week=WeekDay.THURSDAY, year=None, month=None, day=None, hour=Hour.TWELVE_AM,
            minute=0, second=0, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_year_is_set(self):
        """
        Should return a dictionary with
        year set to that year
        month set to Month.January
        day set to 1
        hour set to Hour.TWELVE_AM, and minute, second, millisecond set to 0
        """
        cron_job = CronType(year=1999)
        expected_defaults = dict(
            day_of_week=None, year=1999, month=Month.JANUARY, day=1, hour=Hour.TWELVE_AM,
            minute=0, second=0, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_month_is_set(self):
        """
        Should return a dictionary with
        month set to that month
        day set to 1
        hour set to Hour.TWELVE_AM, and minute, second, millisecond set to 0
        """
        cron_job = CronType(month=Month.MAY)
        expected_defaults = dict(
            day_of_week=None, year=None, month=Month.MAY, day=1, hour=Hour.TWELVE_AM,
            minute=0, second=0, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_day_is_set(self):
        """
        Should return a dictionary with
        day set to that day
        hour set to Hour.TWELVE_AM, and minute, second, millisecond set to 0
        """
        cron_job = CronType(day=4)
        expected_defaults = dict(
            day_of_week=None, year=None, month=None, day=4, hour=Hour.TWELVE_AM,
            minute=0, second=0, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_hour_is_set(self):
        """
        Should return a dictionary with
        hour set to that hour,
        and minute, second, millisecond set to 0
        """
        cron_job = CronType(hour=Hour.TEN_AM)
        expected_defaults = dict(
            day_of_week=None, year=None, month=None, day=None, hour=Hour.TEN_AM,
            minute=0, second=0, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_minute_is_set(self):
        """
        Should return a dictionary with
        minute is set to that minute,
        and second, millisecond set to 0
        """
        cron_job = CronType(minute=45)
        expected_defaults = dict(
            day_of_week=None, year=None, month=None, day=None, hour=None,
            minute=45, second=0, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_second_is_set(self):
        """
        Should return a dictionary with
        second is set to that second,
        millisecond set to 0
        """
        cron_job = CronType(second=44)
        expected_defaults = dict(
            day_of_week=None, year=None, month=None, day=None, hour=None,
            minute=None, second=44, timezone_string='UTC',
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_timezone_string_is_set(self):
        """
        Should return a dictionary with
        timezone_string set to that timezone_string
        """
        cron_job = CronType(timezone_string=self.timezone_name)
        expected_defaults = dict(
            day_of_week=None, year=None, month=None, day=None, hour=None,
            minute=None, second=None, timezone_string=self.timezone_name,
        )
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_generate_field_defaults_when_all_fields_are_set(self):
        """
        Should return a dictionary with all fields appropriately set
        """
        cron_job = CronType(day_of_week=WeekDay.MONDAY, year=1232, month=Month.SEPTEMBER, day=14,
                            hour=Hour.NINE_AM, minute=23, second=10,
                            timezone_string=self.timezone_name)
        expected_defaults = dict(day_of_week=WeekDay.MONDAY, year=1232, month=Month.SEPTEMBER, day=14,
                                 hour=Hour.NINE_AM, minute=23, second=10,
                                 timezone_string=self.timezone_name)
        self.assertDictEqual(cron_job._generate_field_defaults(), expected_defaults)

    def test_get_time(self):
        """
        Should return the time part of the CronType, got from hour, minute, second
        """
        hour = Hour.EIGHT_PM
        minute = 34
        second = 23

        cron_job = CronType(hour=hour, minute=minute, second=second, day_of_week=WeekDay.MONDAY)
        expected_time = time(hour=hour.value, minute=minute, second=second)

        self.assertEqual(expected_time, cron_job._get_time())

    def test_get_date_no_day_of_week(self):
        """
        Should return the date part of the CronType, got from day, month, year
        """
        year = 1998
        month = Month.AUGUST
        day = 22

        cron_job = CronType(year=year, month=month, day=day, hour=Hour.NINE_AM, minute=4)
        expected_date = date(day=day, month=month, year=year)

        self.assertEqual(expected_date, cron_job._get_date())

    @patch('judah.utils.cron.CronType._get_next_date_from_day_of_week')
    def test_get_date_with_day_of_week(self, mock_get_next_date_from_day_of_week):
        """
        Should return the date part of the CronType, got from day_of_week
        """
        expected_date = date(day=3, month=5, year=1994)
        mock_get_next_date_from_day_of_week.return_value = expected_date
        cron_job = CronType(year=1998, day_of_week=WeekDay.SATURDAY)

        self.assertEqual(expected_date, cron_job._get_date())

    def test_get_next_date_from_day_of_week(self):
        """
        Should get the date of the next day_of_week e.g. next Sunday if day_of_week = WeekDay.SUNDAY
        """
        day_of_week = WeekDay.WEDNESDAY
        cron_job = CronType(year=1783, day_of_week=day_of_week)

        today = date.today()
        current_day_of_week = today.weekday()

        if current_day_of_week > day_of_week.value:
            number_of_days_from_today = 7 + day_of_week.value - current_day_of_week
        else:
            number_of_days_from_today = day_of_week.value - current_day_of_week

        expected_date = today + timedelta(days=number_of_days_from_today)
        self.assertEqual(cron_job._get_next_date_from_day_of_week(), expected_date)

    def test_get_next_utc_timestamp_with_day_of_week(self):
        """
        Should return the timestamp for the next day_of_week e.g. the next Sunday
        """
        day_of_week = WeekDay.SUNDAY
        cron_job = CronType(day_of_week=day_of_week)

        today = datetime.now(tz=pytz.utc).astimezone(pytz.UTC)
        current_day_of_week = today.weekday()

        if current_day_of_week >= day_of_week.value:
            number_of_days_from_today = 7 + day_of_week.value - current_day_of_week
        else:
            number_of_days_from_today = day_of_week.value - current_day_of_week

        expected_date = today + timedelta(days=number_of_days_from_today)
        expected_utc_timestamp = datetime(
            year=expected_date.year, month=expected_date.month, day=expected_date.day,
            hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)

        next_timestamp = cron_job.get_next_utc_timestamp()

        self.assertEqual(next_timestamp, expected_utc_timestamp)

    def test_get_next_utc_timestamp_when_only_year_is_set(self):
        """
        Should return the timestamp for 1st January of that year at midnight
        if the timestamp is later than now, else returns None
        """
        today = datetime.utcnow()
        year_in_the_past = 1887
        year_in_future = today.year + 3
        cron_job_in_past_year = CronType(year=year_in_the_past)
        cron_job_in_the_future = CronType(year=year_in_future)

        expected_future_utc_timestamp = datetime(
            year=year_in_future, month=1, day=1,
            hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)
        self.assertEqual(cron_job_in_the_future.get_next_utc_timestamp(), expected_future_utc_timestamp)
        self.assertIsNone(cron_job_in_past_year.get_next_utc_timestamp())

    @freeze_time('1997-04-03 11:46:00')
    def test_get_next_utc_timestamp_when_only_month_is_set(self):
        """
        Should return the timestamp for 1st that month of that year at midnight
        if the timestamp is later than now, else returns 1st Month of the next year
        """
        mock_now = datetime.utcnow()
        current_year = mock_now.year
        month_in_the_past = mock_now.month - 2
        month_in_future = mock_now.month + 3
        cron_job_for_past_month = CronType(month=Month(month_in_the_past))
        cron_job_for_future_month = CronType(month=Month(month_in_future))

        expected_utc_timestamp_for_month_in_past = datetime(
            year=current_year + 1, month=month_in_the_past, day=1,
            hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)
        expected_utc_timestamp_for_month_in_future = datetime(
            year=current_year, month=month_in_future, day=1,
            hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)

        self.assertEqual(
            cron_job_for_past_month.get_next_utc_timestamp(), expected_utc_timestamp_for_month_in_past)
        self.assertEqual(
            cron_job_for_future_month.get_next_utc_timestamp(), expected_utc_timestamp_for_month_in_future)

    @freeze_time('1997-04-13 11:46:00')
    def test_get_next_utc_timestamp_when_only_day_is_set(self):
        """
        Should return the timestamp for that day that month of that year at midnight
        if the timestamp is later than now, else return that day next month
        """
        mock_now = datetime.utcnow()
        current_year = mock_now.year
        current_month = mock_now.month
        day_in_the_past = mock_now.day - 2
        day_in_future = mock_now.day + 3
        cron_job_for_day_in_past = CronType(day=day_in_the_past)
        cron_job_for_day_in_future = CronType(day=day_in_future)

        expected_utc_timestamp_for_day_in_past = datetime(
            year=current_year, month=current_month + 1, day=day_in_the_past,
            hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)
        expected_utc_timestamp_for_day_in_future = datetime(
            year=current_year, month=current_month, day=day_in_future,
            hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)

        self.assertEqual(
            cron_job_for_day_in_past.get_next_utc_timestamp(), expected_utc_timestamp_for_day_in_past)
        self.assertEqual(
            cron_job_for_day_in_future.get_next_utc_timestamp(), expected_utc_timestamp_for_day_in_future)

    @freeze_time('1997-04-13 08:46:00+00:00')
    def test_get_next_utc_timestamp_when_only_hour_is_set(self):
        """
        Should return the timestamp for that hour of that day that month of that year at minute 0, second 0
        if the timestamp is later than now, else return that hour next day
        """
        mock_now = datetime.now(tz=pytz.UTC)
        current_year = mock_now.year
        current_month = mock_now.month
        current_day = mock_now.day

        hour_in_the_past = mock_now.hour - 2
        hour_in_future = mock_now.hour + 3
        cron_job_for_hour_in_past = CronType(hour=Hour(hour_in_the_past))
        cron_job_for_hour_in_future = CronType(hour=Hour(hour_in_future))

        expected_utc_timestamp_for_hour_in_past = datetime(
            year=current_year, month=current_month, day=current_day + 1,
            hour=hour_in_the_past, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)
        expected_utc_timestamp_for_hour_in_future = datetime(
            year=current_year, month=current_month, day=current_day,
            hour=hour_in_future, minute=0, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)

        self.assertEqual(
            cron_job_for_hour_in_past.get_next_utc_timestamp(), expected_utc_timestamp_for_hour_in_past)
        self.assertEqual(
            cron_job_for_hour_in_future.get_next_utc_timestamp(), expected_utc_timestamp_for_hour_in_future)

    @freeze_time('1997-04-13 08:46:00+00:00')
    def test_get_next_utc_timestamp_when_only_minute_is_set(self):
        """
        Should return the timestamp for that minute of that hour of that day that month of that year at second 0
        if the timestamp is later than now, else return that minute next hour
        """
        mock_now = datetime.now(tz=pytz.UTC)
        current_year = mock_now.year
        current_month = mock_now.month
        current_day = mock_now.day
        current_hour = mock_now.hour

        minute_in_the_past = mock_now.minute - 20
        minute_in_future = mock_now.minute + 10
        cron_job_for_minute_in_past = CronType(minute=minute_in_the_past)
        cron_job_for_minute_in_future = CronType(minute=minute_in_future)

        expected_utc_timestamp_for_minute_in_past = datetime(
            year=current_year, month=current_month, day=current_day,
            hour=current_hour + 1, minute=minute_in_the_past, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)
        expected_utc_timestamp_for_minute_in_future = datetime(
            year=current_year, month=current_month, day=current_day,
            hour=current_hour, minute=minute_in_future, second=0, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)

        self.assertEqual(
            cron_job_for_minute_in_past.get_next_utc_timestamp(), expected_utc_timestamp_for_minute_in_past)
        self.assertEqual(
            cron_job_for_minute_in_future.get_next_utc_timestamp(), expected_utc_timestamp_for_minute_in_future)

    @freeze_time('1997-04-13 08:46:30+00:00')
    def test_get_next_utc_timestamp_when_only_second_is_set(self):
        """
        Should return the timestamp for that second of that minute of that hour of that day that month of that year
        if the timestamp is later than now, else return that second the next minute
        """
        mock_now = datetime.now(tz=pytz.UTC)
        current_year = mock_now.year
        current_month = mock_now.month
        current_day = mock_now.day
        current_hour = mock_now.hour
        current_minute = mock_now.minute

        second_in_the_past = mock_now.second - 20
        second_in_future = mock_now.second + 10
        cron_job_for_second_in_past = CronType(second=second_in_the_past)
        cron_job_for_second_in_future = CronType(second=second_in_future)

        expected_utc_timestamp_for_second_in_past = datetime(
            year=current_year, month=current_month, day=current_day,
            hour=current_hour, minute=current_minute + 1, second=second_in_the_past, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)
        expected_utc_timestamp_for_second_in_future = datetime(
            year=current_year, month=current_month, day=current_day,
            hour=current_hour, minute=current_minute, second=second_in_future, microsecond=0, tzinfo=pytz.utc
        ).astimezone(pytz.UTC)

        self.assertEqual(
            cron_job_for_second_in_past.get_next_utc_timestamp(), expected_utc_timestamp_for_second_in_past)
        self.assertEqual(
            cron_job_for_second_in_future.get_next_utc_timestamp(), expected_utc_timestamp_for_second_in_future)


if __name__ == '__main__':
    main()



"""Utility Functions centered around data"""
from typing import Any, Dict


def remove_empty_string_values(value: Any, replacement: Any = None) -> Any:
    """Converts a given empty string to None
    """
    if value == '':
        return replacement

    return value


def replace_keys_with_aliases(alias_map: Dict[str, str], data: Dict[Any, Any]):
    """Returns the data with the right aliases set"""
    return {alias_map.get(key, key): value for key, value in data.items()}


def replace_substrings(substring_replacement_map: Dict[str, str], string: str) -> str:
    """Replaces characters in a given string"""
    new_string = string

    for substring, replacement in substring_replacement_map.items():
        new_string = new_string.replace(substring, replacement)

    return new_string



"""Tests for the datetime transformer"""
from unittest import TestCase, main

from judah.transformers.datetime_transformer import DatetimeTransformer


class TestDatetimeTransformer(TestCase):
    """Tests for the DatetimeTransformer base class"""

    def test_run(self):
        """
        should transform the format of the date field in a given data dictionary
        """
        input_data = {
            'start_date': '09/24/2020 11:45:32',
            'amount': 5789
        }

        expected_output_data = {
            'start_date': '2020-09-24 11-45-32',
            'amount': 5789
        }

        class ChildDatetimeTransformer(DatetimeTransformer):
            datetime_field: str = 'start_date'
            source_datetime_format: str = '%m/%d/%Y %H:%M:%S'
            destination_datetime_format: str = '%Y-%m-%d %H-%M-%S'

        self.assertDictEqual(ChildDatetimeTransformer.run(input_data), expected_output_data)


if __name__ == '__main__':
    main()
